@@ -18,6 +18,16 @@ const patterns = split("""
     *Makefile
 """)
 
+allow_tabs(path) =
+    path == "Make.inc" ||
+    endswith(path, "Makefile") ||
+    endswith(path, ".make") ||
+    endswith(path, ".mk") ||
+    startswith(path, joinpath("src", "support")) ||
+    startswith(path, joinpath("src", "flisp")) ||
+    endswith(path, joinpath("test", "syntax.jl")) ||
+    endswith(path, joinpath("test", "triplequote.jl"))
+
 const errors = Set{Tuple{String,Int,String}}()
 
 for path in eachline(`git ls-files -- $patterns`)
@@ -27,10 +37,13 @@ for path in eachline(`git ls-files -- $patterns`)
     file_err(msg) = push!(errors, (path, 0, msg))
     line_err(msg) = push!(errors, (path, lineno, msg))
 
+    isfile(path) || continue
     for line in eachline(path, keep=true)
         lineno += 1
         contains(line, '\r')   && file_err("non-UNIX line endings")
         contains(line, '\ua0') && line_err("non-breaking space")
+        allow_tabs(path) ||
+        contains(line, '\t')   && line_err("tab")
         endswith(line, '\n')   || line_err("no trailing newline")
         line = chomp(line)
         endswith(line, r"\s")  && line_err("trailing whitespace")
